#include "example-plugin.h"

#include <lib.h>
#include <deliver.h>
#include <array.h>
#include <istream-header-filter.h>
#include <message-header-parser.h>
#include <istream.h>
#include <istream-concat.h>
#include <index/index-mail.h>
#include <istream-crlf.h>


#define PRINT_BANNER i_info("example: func %s is called", __FUNCTION__)

#define MAX_USER_HEADERS 128

#define EXAMPLE_MAIL_CONTEXT(obj) \
	MODULE_CONTEXT((struct mail_private *)obj, example_mail_module)

MODULE_CONTEXT_DEFINE_INIT(example_mail_module, &mail_module_register);

struct example_mail
{
	union mail_module_context module_ctx;

	ARRAY_TYPE(const_string) exclude_headers;
	ARRAY_TYPE(const_string) include_headers;

	unsigned int enabled:1;
};

static deliver_mail_func_t *next_deliver_mail;


static struct istream *
create_stream_for_msgbody(struct istream *stream)
{
	struct istream *tmp = stream;
	struct message_size hdr_size;
	struct message_size body_size;

	PRINT_BANNER;

	i_stream_seek(stream, 0);
	message_get_header_size(stream, &hdr_size, NULL);
	message_get_body_size(stream, &body_size, NULL);

	i_stream_seek(stream, hdr_size.physical_size);
	stream = i_stream_create_limit(stream, body_size.virtual_size);

	i_stream_unref(&tmp);

	return stream;
}


static void example_header_filter_callback(struct message_header_line *hdr,
                                           bool *matched, void *context)
{
	PRINT_BANNER;

	if (hdr && hdr->eoh == TRUE) { *matched = FALSE; }
	else                         { *matched = TRUE;  }

	if (hdr) { i_info("#%s: %s", hdr->name, hdr->value); }
}


static struct istream *
make_header_filter(struct istream *stream, ARRAY_TYPE(const_string) *headers_arr,
                   unsigned int count)
{
	const char *const *headers = NULL;

	PRINT_BANNER;

	if (count) { headers = (const char **)array_idx(headers_arr, 0); }

	return i_stream_create_header_filter(stream,
				             HEADER_FILTER_EXCLUDE |
	                                     HEADER_FILTER_NO_CR |
	                                     HEADER_FILTER_HIDE_BODY,
				             headers,
				             N_ELEMENTS(headers),
				             example_header_filter_callback,
				             NULL);

}


static struct istream *
create_header_stream(ARRAY_TYPE(const_string) *headers_arr, unsigned int count)
{
	unsigned int i = 0;
	struct istream  *result = NULL;
	struct istream **chunks = NULL;
	const char *const *headers = NULL;

	PRINT_BANNER;

	i_assert(count != 0);

	headers = (const char **)array_idx(headers_arr, 0);

	/* create streams for new included headers from headers */
	chunks = t_new(struct istream *, count + 1);
	for (i = 0; i < count; ++i) {
		struct istream *tmp = NULL;

		tmp = i_stream_create_from_data(headers[i], strlen(headers[i]));
		chunks[i] = i_stream_create_lf(tmp);
		i_stream_unref(&tmp);
	}
	chunks[count] = NULL;
	result = i_stream_create_concat(chunks);

	for (i = 0; chunks[i] != NULL; ++i) { i_stream_unref(&chunks[i]); }

	return result;
}


static int example_get_stream(struct mail *mail, struct message_size *hdr_size,
                              struct message_size *body_size,
                              struct istream **stream_r)
{
	struct istream *stream = NULL;
	struct istream **chunks = NULL;
	struct example_mail *m = EXAMPLE_MAIL_CONTEXT(mail);
	const unsigned int max_chunks = 4;

	unsigned int i = 0;
	unsigned int cnt = 0;
	unsigned int exclude_cnt = 0;
	unsigned int include_cnt = 0;

	int ret = 0;

	PRINT_BANNER;

	{/* Test code */
		const char *str = i_strdup("Subject");
		const char *str2 = i_strdup("X-DSPAM: test value\n\n");

		array_append(&m->exclude_headers, &str, 1);
		array_append(&m->include_headers, &str2, 1);
	}

	/* standard get_stream for mail storage format */
	ret = m->module_ctx.super.get_stream(mail, hdr_size,
	                                           body_size, &stream);

	include_cnt = array_count(&m->include_headers);
	exclude_cnt = array_count(&m->exclude_headers);

	if (!m->enabled || (!exclude_cnt && !include_cnt) || ret < 0) {
		*stream_r = stream;
	} else {
		/* create message from chunks with new headers
		   and skip excluded headers */
		chunks = t_new(struct istream *, max_chunks);
		chunks[cnt++] =
			make_header_filter(stream,
		                           &m->exclude_headers, exclude_cnt);
#ifndef WITH_ONLY_HEADERS
		if (include_cnt) {
			chunks[cnt++] =
				create_header_stream(&m->include_headers,
				                     include_cnt);
		}
		chunks[cnt++] = create_stream_for_msgbody(stream);
#endif
		chunks[cnt++] = NULL;

		*stream_r = i_stream_create_concat(chunks);

		/* cleanup */
		for (i = 0; chunks[i] != NULL; ++i) { i_stream_unref(&chunks[i]); }
		i_stream_unref(&stream);
	}

	return ret;
}


static int example_deliver_mail(struct mail_namespace *namespaces,
                                struct mail_storage **storage_r,
                                struct mail *mail,
                                const char *destaddr,
                                const char *mailbox)
{
	enum mail_flags flags = 0;
	const char *const *keywords = NULL;
	struct mail_private *pmail = (struct mail_private *)mail;
	struct example_mail * ex_mail = NULL;

	PRINT_BANNER;

	ex_mail = p_new(pmail->pool, struct example_mail, 1);
	ex_mail->module_ctx.super = pmail->v;
	i_array_init(&ex_mail->exclude_headers, MAX_USER_HEADERS);
	i_array_init(&ex_mail->include_headers, MAX_USER_HEADERS);
	ex_mail->enabled = 0;
	pmail->v.get_stream = example_get_stream;

	MODULE_CONTEXT_SET(pmail, example_mail_module, ex_mail);

	ex_mail->enabled = 1;
	(void)deliver_save(namespaces, storage_r, mailbox, mail, flags, keywords);
	ex_mail->enabled = 0;
	array_clear(&ex_mail->include_headers);
	array_clear(&ex_mail->exclude_headers);

	return 1;
}


void example_plugin_init(void)
{
	PRINT_BANNER;

	next_deliver_mail = deliver_mail;
	deliver_mail = example_deliver_mail;
}


void example_plugin_deinit(void)
{
	PRINT_BANNER;

	deliver_mail = next_deliver_mail;
}


